from __future__ import annotations

import multiprocessing as mp


class MPool:
    """A pool of processes assigned to a single target function"""

    def __init__(self, num_processes: int, target: callable) -> None:
        """Create a collection of processes assigned to execute a given callable

        Args:
            num_processes: The number of processes to allocate
            target: The function to be executed by the allocated processes
        """

        if num_processes < 0:
            raise ValueError(f'Cannot instantiate negative forked processes (got {num_processes}).')

        # Note that we use the memory address of the processes and not the
        # ``pid`` attribute. ``pid`` is only set after the process is started.
        self._processes = [mp.Process(target=self._call_target) for _ in range(num_processes)]
        self._states = mp.Manager().dict({id(p): False for p in self._processes})
        self._target = target

    def _call_target(self) -> None:  # pragma: nocover, Called from forked process
        """Wrapper for calling the pool's target function"""

        self._target()

        # Mark the current process as being finished
        self._states[id(mp.current_process())] = True

    @property
    def num_processes(self) -> int:
        """The number of processes assigned to the pool"""

        return len(self._processes)

    @property
    def target(self) -> callable:
        """The callable to be executed by the pool"""

        return self._target

    def is_finished(self) -> bool:
        """Return whether all processes have finished executing"""

        # Check that all forked processes are finished
        return all(self._states.values())

    def _raise_if_zero(self, action):
        """Raise an error if pool size is zero"""

        if self.num_processes == 0:
            raise RuntimeError(f'Pool has zero assigned processes. No processes available to {action}')

    def start(self) -> None:
        """Start all processes asynchronously"""

        self._raise_if_zero('start')
        for p in self._processes:
            p.start()

    def join(self) -> None:
        """Wait for any running pool processes to finish running before continuing execution"""

        self._raise_if_zero('join')
        if self.num_processes == 0:
            raise RuntimeError('Pool has zero assigned processes. No processes available to join')

        for p in self._processes:
            p.join()

    def kill(self) -> None:
        """Kill all running processes without trying to exit gracefully"""

        self._raise_if_zero('kill')
        for p in self._processes:
            p.terminate()



"""Callbacks used to update the content of Dash components"""

from __future__ import annotations

from datetime import datetime
from typing import List, Optional, TYPE_CHECKING, Tuple

import numpy as np
import psutil

if TYPE_CHECKING:  # pragma: no cover
    from egon.connectors import Input


def cast_layout_to_dict(layout: str) -> dict:
    """Return a cytoscape layout as a dictionary

    Args:
        layout: The layout of the cytoscape figure

    Returns:
        A dictionary for styling the cytoscape object
    """

    return {'name': layout, 'animate': True}


class BaseGraphCallback:
    """Base class for graph callbacks"""

    def __init__(self, max_data: Optional[int]) -> None:
        """A customized Dash graph

        Args:
            max_data: Maximum number of data points to display on the graph
        """

        self.max_data = max_data


class SystemUsage(BaseGraphCallback):
    """Callbacks for the current system usage"""

    # noinspection PyUnusedLocal
    def get_cpu_usage(self, *args: Optional) -> Tuple[dict, None, int]:
        """Return a dictionary with the current CPU usage

        Returns:
            A dictionary of CPU % usage formatted for use with plotly
        """

        now = datetime.now()
        y = np.transpose([psutil.cpu_percent(percpu=True)])
        x = np.full_like(y, now, dtype=datetime)
        return dict(x=x, y=y), None, self.max_data

    # noinspection PyUnusedLocal
    def get_memory_usage(self, *args: Optional) -> Tuple[dict, None, int]:
        """Return a dictionary with the current memory usage

        Returns:
            A dictionary of memory usage formatted for use with plotly
        """

        mem = psutil.virtual_memory()[2]
        return dict(x=[[datetime.now()]], y=[[mem]]), None, self.max_data


class PipelineStatus(BaseGraphCallback):
    """Callbacks for the status of a pipeline"""

    def __init__(self, max_data: int, connector_list: List[Input]):
        """A customized Dash graph

        Args:
            max_data: Maximum number of data points to display on the graph
            connector_list: A list of input connectors
        """

        super().__init__(max_data=max_data)
        self._connector_list = connector_list

    # noinspection PyUnusedLocal
    def get_queue_sizes(self, *args: Optional) -> Tuple[dict, None, int]:
        """Return a dictionary with the que size for a list of connectors

        Returns:
            A dictionary of queue sizes formatted for use with plotly
        """

        now = datetime.now()
        y = [[c.size()] for c in self._connector_list]
        x = np.full_like(y, now, dtype=datetime)
        return dict(x=x, y=y), None, self.max_data



"""The ``mock`` module defines prebuilt pipeline nodes for developing
unittests. Instead of accomplishing a user defined action, mock nodes sleep
for a pre-defined number of seconds.
"""

from abc import ABC
from typing import Optional

from egon import nodes
from egon.connectors import Input, Output
from egon.nodes import AbstractNode


class MockObject(AbstractNode, ABC):
    """Base class for mock testing nodes"""

    def __init__(self, name: Optional[str] = None) -> None:
        self._is_finished = False
        super(MockObject, self).__init__(name=name)

    def is_finished(self) -> bool:
        """Return whether the mock node has already been executed"""

        return self._is_finished

    def execute(self) -> None:
        """Execute the mock pipeline node"""

        super(MockObject, self)._target()
        self._is_finished = True


class MockSource(MockObject, nodes.Source):
    """A ``Source`` subclass that implements placeholder functions for abstract methods"""

    def __init__(self, load_data: list = None, name: Optional[str] = None) -> None:
        self.output = Output()
        self.load_data = load_data or []
        super(MockSource, self).__init__(name)

    def action(self) -> None:
        """Placeholder function to satisfy requirements of abstract parent"""

        for x in self.load_data:
            self.output.put(x)


class MockTarget(MockObject, nodes.Target):
    """A ``Target`` subclass that implements placeholder functions for abstract methods"""

    def __init__(self, name: Optional[str] = None) -> None:
        self.input = Input()
        self.accumulated_data = []
        super(MockTarget, self).__init__(name)

    def action(self) -> None:
        """Placeholder function to satisfy requirements of abstract parent"""

        for x in self.input.iter_get():
            self.accumulated_data.append(x)


class MockNode(MockObject, nodes.Node):
    """A ``Node`` subclass that implements placeholder functions for abstract methods"""

    def __init__(self, name: Optional[str] = None) -> None:
        self.output = Output()
        self.input = Input()
        super(MockNode, self).__init__(name)

    def action(self) -> None:  # pragma: no cover
        """Placeholder function to satisfy requirements of abstract parent"""

        for x in self.input.iter_get():
            self.output.put(x)



"""Tests for the connection of multiple inputs and outputs together"""

from time import sleep
from unittest import TestCase

from egon.connectors import Input, Output
from egon.mock import MockSource, MockTarget


class MultiplePartnerMapping(TestCase):
    """Test connectors with an established connection correctly map to neighboring connectors/nodes"""

    def setUp(self) -> None:
        """Connect two outputs to a single input"""

        self.input = Input()
        self.output1 = Output()
        self.output2 = Output()

        self.output1.connect(self.input)
        self.output2.connect(self.input)

    def test_output_maps_to_partners(self) -> None:
        """Test connectors map to the correct partner connector"""

        output_connectors = [self.output1, self.output2]
        self.assertCountEqual(output_connectors, self.input.partners)

    def test_input_maps_to_partners(self) -> None:
        """Test connectors map to the correct partner connector"""

        input_connectors = [self.input]
        self.assertCountEqual(input_connectors, self.output1.partners)
        self.assertCountEqual(input_connectors, self.output2.partners)


class MultiplePartnerDataRouting(TestCase):
    """Test output connectors support sending data to multiple input connectors"""

    def runTest(self) -> None:
        # Create one node to output data and two to accept it
        test_data = [1, 2, 3]
        source = MockSource(test_data)
        target_a = MockTarget()
        target_b = MockTarget()

        # Connect two outputs to the same input
        source.output.connect(target_a.input)
        source.output.connect(target_b.input)

        source.execute()
        sleep(1)  # Give the queue a chance to update

        # Both inputs should have received the same data from the output
        target_a.execute()
        self.assertListEqual(test_data, target_a.accumulated_data)

        target_b.execute()
        self.assertListEqual(test_data, target_b.accumulated_data)



"""The Pipeline module defines the ``Pipeline`` class which is responsible
for representing collections of interconnected analysis steps (``Node``
objects) as a  single, coherent analysis pipeline. ``Pipeline`` instances
are also responsible for starting/stopping forked processes and validating
that nodes are properly interconnected.
"""

from __future__ import annotations

import os
import warnings
from asyncio.subprocess import Process
from copy import copy
from inspect import getmembers
from itertools import chain
from typing import List, Tuple

from . import connectors as conn
from . import nodes
from .visualize import Visualizer


class Pipeline:
    """Manages a collection of nodes as a single analysis pipeline"""

    def __init__(self) -> None:

        # Store the nodes and connectors used to build the pipeline
        # so they can be exposed by public accessors
        self._inputs: List[conn.Input] = []
        self._outputs: List[conn.Output] = []
        self._sources: List[nodes.Source] = []
        self._inlines: List[nodes.Node] = []
        self._targets: List[nodes.Target] = []

        for attr_name, *_ in getmembers(self, lambda a: isinstance(a, nodes.AbstractNode)):
            node = getattr(self, attr_name)
            if isinstance(node, nodes.Node):
                self._inlines.append(node)

            elif isinstance(node, nodes.Source):
                self._sources.append(node)

            else:  # Assume all other nodes are inlines
                self._targets.append(node)

            inputs, outputs = node.connectors
            self._inputs.extend(inputs)
            self._outputs.extend(outputs)

    def validate(self) -> None:
        """Set up the pipeline and check for any invalid node states"""

        for node in chain(*self.nodes):
            node.validate()

    def _get_processes(self) -> List[Process]:
        """Return a list of processes forked by pipeline nodes"""

        # Collect all of the processes assigned to each node
        processes = []
        for node in chain(*self.nodes):
            processes.extend(node._pool._processes)

        return processes

    @property
    def nodes(self) -> Tuple[Tuple[nodes.Source], Tuple[nodes.Node], Tuple[nodes.Target]]:
        """Return a list of all nodes in the pipeline

        Nodes are returned in an arbitrary order

        Returns:
            A list of nodes used to build the pipeline
        """

        return tuple(self._sources), tuple(self._inlines), tuple(self._targets)

    @property
    def connectors(self) -> Tuple[List[conn.Input], List[conn.Output]]:
        """Return the input and output connectors used by the pipeline

        Returns:
            A tuple with a list of input connectors and a list of output connectors
        """

        return copy(self._inputs), copy(self._outputs)

    @property
    def num_processes(self) -> int:
        """The number of processes forked by to the pipeline"""

        return len(self._get_processes())

    def kill(self) -> None:
        """Kill all running pipeline processes without trying to exit gracefully"""

        for p in self._get_processes():
            p.terminate()

    def run(self) -> None:
        """Start all pipeline processes and block execution until all processes exit"""

        self.run_async()
        self.wait_for_exit()

    def wait_for_exit(self) -> None:
        """Wait for the pipeline to finish running before continuing execution"""

        for node in chain(*self.nodes):
            node._pool.join()

    def run_async(self) -> None:
        """Start all processes asynchronously"""

        for node in chain(*self.nodes):
            node._pool.start()

    def visualize(
            self,
            host: str = os.getenv("EGON_HOST", "127.0.0.1"),
            port: int = os.getenv("EGON_PORT", "8050"),
            quiet: bool = False
    ) -> None:
        """Launch a server instance for monitoring the pipeline in real time

        This is a blocking call intended to be run in the main application
        process.

        Args:
            host: Host IP used to serve the application
            port: Port used to serve the application
            quiet: Suppress any status messages or warnings
        """

        from waitress import serve

        if not quiet:
            print(f'Launching server at http://{host}:{port}')

        # we increase the number of threads from 4 (the default) to 8 so the
        # server can more efficiently handle application callbacks
        with warnings.catch_warnings():
            if quiet:
                warnings.simplefilter('ignore')

            serve(Visualizer(self).server, host=host, port=int(port), threads=8, _quiet=quiet)



"""The ``nodes`` module supports the construction of individual pipeline nodes.
``Source``, ``Node``,  and ``Target`` classes are provided for creating nodes
that produce, analyze, and consume data respectively.
"""

from __future__ import annotations

import abc
from abc import ABC
from itertools import chain
from typing import Collection
from typing import List, Tuple, Union

from . import connectors, exceptions
from .multiprocessing import MPool


def _get_nodes_from_connectors(connector_list: Collection[connectors.BaseConnector]) -> Tuple:
    """Return the parent nodes from a list of ``Connector`` objects
    
    Args:
        connector_list: The connectors to get parents of
        
    Returns:
        A list of node instances
    """

    return tuple(p.parent_node for c in connector_list for p in c.partners)


class AbstractNode(abc.ABC):
    """Base class for constructing pipeline nodes"""

    def __init__(self, name: str = None, num_processes: int = 1) -> None:
        """Represents a single pipeline node"""

        self._pool: MPool = MPool(num_processes, self._target)
        self._allow_pool_overwrite = True  # See setter for ``num_processes`` attribute
        self.name = name or self.__class__.__name__

        # Accumulate all attributes that are Input or Output types
        self._inputs = []
        self._outputs = []
        for connector in self._get_attrs(connectors.BaseConnector):
            connector._node = self  # Make the connector aware of its parent node
            if isinstance(connector, connectors.Input):
                self._inputs.append(connector)

            else:  # Assume all other connectors are outputs
                self._outputs.append(connector)

        self._inputs = tuple(self._inputs)
        self._outputs = tuple(self._outputs)

    def _get_attrs(self, attr_type=None) -> List:
        """Return a list of instance attributes matching the given type

        All class attributes are ignored.

        Args:
            attr_type: The object type to search for

        Returns:
            A list of attributes with type ``attr_type``
        """

        attr_list = []
        ignore = dir(self.__class__)

        for attr_name in dir(self):
            if (not attr_name.startswith('_')) and (attr_name not in ignore):
                attr = getattr(self, attr_name)
                if isinstance(attr, attr_type):
                    attr_list.append(attr)

        return attr_list

    @property
    def num_processes(self) -> int:
        """The number of processes assigned to the pool"""

        return self._pool.num_processes

    @num_processes.setter
    def num_processes(self, val) -> None:
        if not self._allow_pool_overwrite:
            raise RuntimeError('Cannot change number of processes on running or finished node')

        self._pool = MPool(val, self._target)

    @property
    def connectors(self) -> Tuple[Tuple[connectors.Input, ...], Tuple[connectors.Output, ...]]:
        """Return tuples with all input and output connectors associated with the node

        Returns:
            A tuple of Input connectors and a tuple of Output connectors
        """

        return self._inputs, self._outputs

    @property
    def upstream_nodes(self) -> Tuple[Union[Source, Node]]:
        """Returns a list of nodes that are upstream from the current node"""

        return _get_nodes_from_connectors(self._get_attrs(connectors.Input))

    @property
    def downstream_nodes(self) -> Tuple[Union[Node, Target]]:
        """Returns a list of nodes that are downstream from the current node"""

        return _get_nodes_from_connectors(self._get_attrs(connectors.Output))

    @abc.abstractmethod
    def validate(self) -> None:
        """Raise an exception if the node object was constructed improperly

        Raises:
            ValueError: For an invalid instance construction
        """

    def _validate_connections(self) -> None:
        """Raise an exception if any of the node's Inputs/Outputs are missing connections

        Raises:
            MissingConnectionError: For an invalid instance construction
        """

        for conn in chain(*self.connectors):
            if not conn.is_connected():
                raise exceptions.MissingConnectionError(
                    f'Connector {conn} does not have an established connection (Node: {conn.parent_node})')

    @abc.abstractmethod
    def action(self) -> None:
        """The primary analysis task performed by this node"""

    def setup(self) -> None:
        """Setup tasks called before running ``action``"""

    def teardown(self) -> None:
        """Teardown tasks called after running ``action``"""

    def _target(self) -> None:
        """Entry point for running the node in a multiprocessing pool"""

        self._allow_pool_overwrite = False
        self.setup()
        self.action()
        self.teardown()

    def execute(self):
        """Execute the pipeline node

        Execution includes all ``setup``, ``action``, and ``teardown`` tasks.
        """

        self._pool.start()
        self._pool.join()

    def is_finished(self) -> bool:
        """Return whether all node processes have finished processing data

        The returned value defaults to ``True`` when the number of processes
        assigned to the node instance is zero.
        """

        return self._pool.is_finished()

    def is_expecting_data(self) -> bool:
        """Return whether the node is still expecting data from upstream

        This function includes checks for whether any upstream nodes are still
        running or any data is pending in the queue of an input connector.
        """

        for input_connector in self._get_attrs(connectors.Input):
            # IMPORTANT: The order of the following code blocks is crucial
            # We check for any running upstream nodes first
            for output_connector in input_connector.partners:
                if not output_connector.parent_node.is_finished():
                    return True

            # Check for any unprocessed data once we know there are no
            # nodes still populating any input queues
            if not input_connector.is_empty():
                return True

        return False

    def __str__(self) -> str:  # pragma: no cover
        return f'<{self.__repr__()} object at {hex(id(self))}>'

    def __repr__(self) -> str:  # pragma: no cover
        return f'{self.__class__.__name__}(num_processes={self.num_processes})'


class Source(AbstractNode, ABC):
    """A pipeline process that only has output streams"""

    def validate(self) -> None:
        """Raise exception if the object is not a valid instance

        Raises:
            MalformedSourceError: For an invalid instance construction
            OrphanedNodeError: For an instance that is inaccessible by connectors
        """

        if self._get_attrs(connectors.Input):
            raise exceptions.MalformedSourceError('Source objects cannot have upstream components.')

        if not self._get_attrs(connectors.Output):
            raise exceptions.OrphanedNodeError('Source has no output connectors and is inaccessible by the pipeline.')

        self._validate_connections()


class Target(AbstractNode, ABC):
    """A pipeline process that only has input streams"""

    def validate(self) -> None:
        """Raise exception if the object is not a valid instance

        Raises:
            MalformedTargetError: For an invalid instance construction
            OrphanedNodeError: For an instance that is inaccessible by connectors
        """

        if self._get_attrs(connectors.Output):
            raise exceptions.MalformedTargetError('Source objects cannot have upstream components.')

        if not self._get_attrs(connectors.Input):
            raise exceptions.OrphanedNodeError('Target has no input connectors and is inaccessible by the pipeline.')

        self._validate_connections()


class Node(AbstractNode, ABC):
    """A pipeline process that can have any number of input or output streams"""

    def validate(self) -> None:
        """Raise exception if the object is not a valid instance

        Raises:
            OrphanedNodeError: For an instance that is inaccessible by connectors
        """

        inputs, outputs = self.connectors
        if not (inputs or outputs):
            raise exceptions.OrphanedNodeError('Node has no associated connectors and is inaccessible by the pipeline.')

        self._validate_connections()



"""Launches a web app for visualizing the status of a pipeline"""

from __future__ import annotations

from typing import TYPE_CHECKING

import dash
import dash.dependencies as ddep
from dash import dcc
from dash import html as dhtml

from egon.visualize import callbacks
from egon.visualize import components as ecomp

if TYPE_CHECKING:  # pragma: no cover
    from egon.pipeline import Pipeline

FORMAT_LABELS = ('Grid', 'Breadth First', 'Circle')
CYTOSCAPE_FORMATS = ('grid', 'breadthfirst', 'circle')
DEFAULT_LAYOUT = 'grid'


class Visualizer(dash.Dash):
    """Application for visualizing an analysis pipeline"""

    def __init__(self, pipeline: Pipeline, update_interval: int = 1, display_interval: int = 120) -> None:
        """Create an interactive application for viewing pipeline objects

        Args:
            pipeline: The pipeline to build the application around
        """

        pipeline.validate()

        self._pipeline = pipeline
        self._update_interval = update_interval
        self._num_displayed_data_points = display_interval // update_interval

        # self.layout = self._build_html()
        super().__init__(__name__, update_title=None, title='Egon Visualizer')
        self._assign_callbacks()

    def _assign_callbacks(self) -> None:
        """Assign callbacks to connect templated HTML with pipeline behavior"""

        self.callback(
            ddep.Output('pipeline-cyto', 'layout'),
            ddep.Input('dropdown-layout', 'value')
        )(callbacks.cast_layout_to_dict)

        pipeline_callbacks = callbacks.PipelineStatus(self._num_displayed_data_points, self._pipeline.connectors[0])
        self.callback(
            ddep.Output('graph-queue-size', 'extendData'),
            ddep.Input('interval', 'n_intervals')
        )(pipeline_callbacks.get_queue_sizes)

        system_callbacks = callbacks.SystemUsage(self._num_displayed_data_points)
        self.callback(
            ddep.Output('graph-cpu-usage', 'extendData'), ddep.Input('interval', 'n_intervals')
        )(system_callbacks.get_cpu_usage)

        self.callback(
            ddep.Output('graph-mem-usage', 'extendData'), ddep.Input('interval', 'n_intervals')
        )(system_callbacks.get_memory_usage)

    #################################################################################################
    # Some of the custom graph objects populate a placeholder value at init
    # By overwriting the _layout attribute, graphs are re-rendered every time the webpage refreshes.
    #################################################################################################

    @property
    def _layout(self) -> dhtml.Div:
        """Create the HTML content to be displayed by the app

        Returns:
            An HTML component
        """

        update_interval_ms = self._update_interval * 1000  # The interval in milliseconds
        page_update_interval = dcc.Interval(id='interval', interval=update_interval_ms)

        # The page has left and right columns which we build separately
        # Here we start building the left column
        left_column = \
            dhtml.Div(className='three columns div-left-panel', children=[
                ecomp.custom.Logo(),
                dcc.Markdown(
                    'This interface provides a general pipeline overview for testing and debugging purposes. '
                    'Please consult with a system administrator before running on a cluster environment. '
                    f'For more information see the [official documentation]({ecomp.custom.DOC_URL}).'
                ),
                dhtml.P(
                    'The structural preview of your pipeline will load with an automatically optimized '
                    'layout. Depending on the number of nodes in your pipeline and the size of your '
                    'window, you may need to manually reposition plot elements by dragging them around '
                    'the window. Predefined layouts are also provided via the dropdown box at the top '
                    'of the page.'
                ),
                dhtml.H6('Pipeline Summary', id='h6-summary'),
                ecomp.custom.SummaryTable(self._pipeline)
            ])

        # Here we start building the right column
        dropdown_layout_selector = \
            dhtml.Div(className='div-dropdown-layout', children=[
                dcc.Dropdown(
                    id='dropdown-layout',
                    value=DEFAULT_LAYOUT,
                    clearable=False,
                    options=[{'label': label, 'value': name} for name, label in zip(CYTOSCAPE_FORMATS, FORMAT_LABELS)]
                )
            ])

        right_column = \
            dhtml.Div(className='nine columns div-right-panel', children=[
                dropdown_layout_selector,
                ecomp.cyto.PipelineCytoscape(self._pipeline, id='pipeline-cyto'),
                dhtml.H4('Pipeline Load'),
                dhtml.Div(
                    className='div-pipeline-load',
                    children=[
                        ecomp.graphs.PipelineQueueSize(self._pipeline, id='graph-queue-size')
                    ]),
                dhtml.H4('System Load'),
                dhtml.Div(
                    className='div-system-load',
                    children=[
                        ecomp.graphs.CpuPercentageGraph(id='graph-cpu-usage'),
                        ecomp.graphs.RamPercentageGraph(id='graph-mem-usage')
                    ]),
            ])

        # Return the fully assembled page with both columns
        return dhtml.Div(className='row', children=[left_column, right_column, page_update_interval])

    @_layout.setter
    def _layout(self, x):
        pass



"""Custom Dash components that are highly specific to the Egon package."""

from dash import html as dhtml

DOC_URL = 'https://mwvgroup.github.io/Egon/'


class Logo(dhtml.Div):
    """The Egon application logo"""

    def __init__(self) -> None:
        super().__init__(className='div-logo', children=[
            dhtml.A(target=DOC_URL, className='display-inline', children=[
                dhtml.Img(className='logo', src='https://mwvgroup.github.io/Egon/assets/images/logo.svg'),
                dhtml.Span(className='logo-text', children=['Egon'])
            ])
        ])


class SummaryTable(dhtml.Div):
    """Summary table for pipeline metrics"""

    def __init__(self, pipeline) -> None:
        inputs, outputs = pipeline.connectors
        number_inputs = len(inputs)
        number_outputs = len(outputs)
        total_connectors = number_inputs + number_outputs

        sources, nodes, targets = pipeline.nodes
        num_sources = len(sources)
        num_nodes = len(nodes)
        num_targets = len(targets)
        number_nodes = num_sources + num_nodes + num_targets

        super().__init__(className='div-info-table', children=[
            dhtml.Table(className='table-info-primary', children=[
                dhtml.Tr(children=[
                    dhtml.Td(children=[f'Total Nodes: {number_nodes}'])]),
                dhtml.Tr(children=[
                    dhtml.Td(children=[
                        dhtml.Table(className='table-info-secondary', children=[
                            dhtml.Tr(children=[
                                dhtml.Td(children=['Source:']),
                                dhtml.Td(children=[f'{num_sources}'])]),
                            dhtml.Tr(children=[
                                dhtml.Td(children=['Inline:']),
                                dhtml.Td(children=[f'{num_nodes}'])]),
                            dhtml.Tr(children=[
                                dhtml.Td(children=['Target:']),
                                dhtml.Td(children=[f'{num_targets}'])
                            ])])])]),

                dhtml.Tr(children=[
                    dhtml.Td(children=[f'Total Connectors: {total_connectors}'])]),
                dhtml.Tr(children=[
                    dhtml.Td(children=[
                        dhtml.Table(className='table-info-secondary', children=[
                            dhtml.Tr(children=[
                                dhtml.Td(children=['Input:']),
                                dhtml.Td(children=[f'{number_inputs}'])]),
                            dhtml.Tr(children=[
                                dhtml.Td(children=['Output:']),
                                dhtml.Td(children=[f'{number_outputs}'])])
                        ])])])
            ])
        ])



"""Tests for the class based construction of pipeline nodes."""

from functools import partial
from time import sleep
from unittest import TestCase

from egon import mock
from egon.exceptions import MissingConnectionError
from egon.mock import MockSource


class Execution(TestCase):
    """Test the execution of tasks assigned to a Node instance"""

    def setUp(self) -> None:
        """Create a testing node that tracks the execution method of it's methods"""

        self.node = mock.MockNode()

        # Track the call order of node functions
        self.call_list = []
        self.node.setup = partial(self.call_list.append, 'setup')
        self.node.action = partial(self.call_list.append, 'action')
        self.node.teardown = partial(self.call_list.append, 'teardown')

    def test_call_order(self) -> None:
        """Test that setup and teardown actions are called in the correct order"""

        self.node.execute()
        self.assertListEqual(self.call_list, ['setup', 'action', 'teardown'])


class TreeNavigation(TestCase):
    """Test ``Node`` instances are aware of their neighbors"""

    def setUp(self) -> None:
        """Create a tree of ``MockNode`` instances"""

        self.root = mock.MockSource()
        self.inline = mock.MockNode()
        self.target = mock.MockTarget()

        self.root.output.connect(self.inline.input)
        self.inline.output.connect(self.target.input)

    def test_upstream_nodes(self) -> None:
        """Test the inline node resolves the correct parent node"""

        self.assertEqual(self.root, self.inline.upstream_nodes[0])
        self.assertEqual(self.inline, self.target.upstream_nodes[0])

    def test_downstream_nodes(self) -> None:
        """Test the inline node resolves the correct child node"""

        self.assertEqual(self.inline, self.root.downstream_nodes[0])
        self.assertEqual(self.target, self.inline.downstream_nodes[0])


class ExpectingData(TestCase):
    """Tests for the ``is_expecting_data`` function

    The ``is_expecting_data`` function combines two booleans.
    This class evaluates all four squares of the corresponding truth table
    """

    def setUp(self) -> None:
        """Create a tree of ``MockNode`` instances"""

        self.root = mock.MockSource()
        self.node = mock.MockNode()
        self.root.output.connect(self.node.input)

    def test_false_for_empty_queue_and_finished_parent(self) -> None:
        """Test the return is False for a EMPTY queue and a FINISHED PARENT node"""

        self.root.execute()
        self.assertFalse(self.node.is_expecting_data())

    def test_true_if_input_queue_has_data(self) -> None:
        """Test the return is True for a NOT EMPTY queue and a FINISHED PARENT node"""

        self.root.execute()
        self.node.input._queue.put(5)
        sleep(1)  # Give the queue an opportunity to update

        self.assertTrue(self.node.is_expecting_data())

    def test_true_if_parent_is_running(self) -> None:
        """Test the return is True for a EMPTY queue and a NOT FINISHED PARENT node"""

        self.assertTrue(self.node.is_expecting_data())

    def test_true_if_input_queue_has_data_and_parent_is_running(self) -> None:
        """Test the return is True for a NOT EMPTY queue and a NOT FINISHED PARENT node"""

        self.node.input._queue.put(5)
        self.assertTrue(self.node.is_expecting_data())


class ConnectorValidation(TestCase):
    """Test the validation of underlying connector objects"""

    def test_missing_connection(self) -> None:
        """Test a ``MissingConnectionError`` for an unconnected connector"""

        with self.assertRaises(MissingConnectionError):
            MockSource().validate()



"""Build a simple two pipeline and test all input data makes it
through to the end.
"""

from multiprocessing import Queue
from unittest import TestCase

from egon.decorators import as_node, as_source, as_target
from egon.pipeline import Pipeline

TESTING_VALS = list(range(10))  # Input values for the pipeline
GLOBAL_QUEUE = Queue()  # For storing pipeline outputs


@as_source
def sending_node() -> float:
    """Load data into the pipeline"""

    for i in TESTING_VALS:
        yield i


@as_node
def internal_node(x) -> float:
    """Returns the input argument"""

    return x


@as_target
def receiving_node(x) -> None:
    """Retrieve data out of the pipeline"""

    GLOBAL_QUEUE.put(x)


class AddingPipeline(Pipeline):
    """A pipeline for generating and then adding numbers"""

    def __init__(self) -> None:
        self.send_node = sending_node
        self.internal_node = internal_node
        self.receive_node = receiving_node

        self.send_node.output.connect(self.internal_node.input)
        self.internal_node.output.connect(self.receive_node.input)

        super().__init__()
        self.validate()


class TestPipelineThroughput(TestCase):
    """Test all data makes it through the pipeline"""

    def runTest(self) -> None:
        """Compare input and output pipeline values"""

        # Run should populate the global queue
        AddingPipeline().run()

        # Convert the queue into a list
        as_list = []
        while GLOBAL_QUEUE.qsize() != 0:
            as_list.append(GLOBAL_QUEUE.get())

        self.assertCountEqual(TESTING_VALS, as_list)



"""Tests the functionality of ``Output`` connector objects."""

from unittest import TestCase

from egon import exceptions
from egon.connectors import Input, Output
from egon.exceptions import MissingConnectionError


class DataPut(TestCase):
    """Test data routing by ``Output`` instances"""

    def test_stores_value_in_queue(self) -> None:
        """Test the ``put`` method puts data into the underlying connected queue"""

        # Create a node with an output connector
        output = Output()
        input = Input()
        output.connect(input)

        test_val = 'test_val'
        output.put(test_val)
        self.assertEqual(input._queue.get(), test_val)

    def test_error_if_unconnected(self) -> None:
        """Test ``put`` raises an error if the output is not connected"""

        with self.assertRaises(MissingConnectionError):
            Output().put(5)

    @staticmethod
    def test_error_override() -> None:
        """Test the optional suppression of errors for unconnected outputs"""

        Output().put(5, raise_missing_connection=False)


class InstanceConnect(TestCase):
    """Test the connection of generic connector objects to other"""

    def setUp(self) -> None:

        self.input_connector = Input()
        self.output_connector = Output()

    def test_error_on_connection_to_same_type(self) -> None:
        """Test an error is raised when connecting two outputs together"""

        with self.assertRaises(ValueError):
            self.output_connector.connect(Output())

    def test_overwrite_error_on_connection_overwrite(self) -> None:
        """Test an error is raised when trying to overwrite an existing connection"""

        input = Input()
        self.output_connector.connect(input)
        with self.assertRaises(exceptions.OverwriteConnectionError):
            self.output_connector.connect(input)


class InstanceDisconnect(TestCase):
    """Test the disconnection of two connectors"""

    def setUp(self) -> None:
        self.input = Input()
        self.output = Output()
        self.output.connect(self.input)

    def test_both_connectors_are_disconnected(self) -> None:
        """Test both connectors are no longer listed as partners"""

        self.assertTrue(self.input.is_connected())
        self.assertTrue(self.output.is_connected())

        self.output.disconnect(self.input)
        self.assertNotIn(self.output, self.input.partners)
        self.assertFalse(self.input.is_connected())
        self.assertFalse(self.output.is_connected())

    def test_error_if_not_connected(self) -> None:
        """Test an error is raised when disconnecting a connector that is not connected"""

        with self.assertRaises(MissingConnectionError):
            Output().disconnect(Input())



"""Tests for the ``Pipeline`` class"""

from unittest import TestCase

from egon.exceptions import MissingConnectionError, OrphanedNodeError
from egon.mock import MockNode, MockSource, MockTarget
from egon.nodes import Node
from egon.pipeline import Pipeline


class SimplePipeline(Pipeline):
    """Basic pipeline for testing the parent pipeline class"""

    def __init__(self) -> None:
        self.source = MockSource([1, 2, 3])
        self.inline = MockNode()
        self.target = MockTarget()

        self.source.output.connect(self.inline.input)
        self.inline.output.connect(self.target.input)
        super(SimplePipeline, self).__init__()


class ProcessDiscovery(TestCase):
    """Test the pipeline is aware of all processes forked by it's nodes"""

    @classmethod
    def setUpClass(cls) -> None:
        cls.pipeline = SimplePipeline()
        cls.expected_processes = []
        cls.expected_processes.extend(cls.pipeline.source._pool._processes)
        cls.expected_processes.extend(cls.pipeline.inline._pool._processes)
        cls.expected_processes.extend(cls.pipeline.target._pool._processes)

    def test_collected_processes_match_nodes(self) -> None:
        """Test ``_get_processes`` returns forked processes from all pipeline nodes"""

        self.assertCountEqual(self.expected_processes, self.pipeline._get_processes())

    def test_process_count(self) -> None:
        """Test the pipelines process count matches the sum of processes allocated to each node"""

        self.assertEqual(len(self.expected_processes), self.pipeline.num_processes)


class NodeDiscovery(TestCase):
    """Test the pipeline is aware of all it's nodes"""

    def runTest(self) -> None:
        """Instantiate a test pipeline and ask it for it's available nodes"""

        pipeline = SimplePipeline()
        sources, inlines, targets = pipeline.nodes

        self.assertEqual((pipeline.source,), sources)
        self.assertEqual((pipeline.inline,), inlines)
        self.assertEqual((pipeline.target,), targets)


class ConnectorDiscovery(TestCase):
    """Test the pipeline is aware of all it's connector objects"""

    def runTest(self) -> None:
        """Instantiate a test pipeline and ask it for it's available nodes"""

        pipeline = SimplePipeline()
        expected_inputs = []
        expected_inputs.extend(pipeline.source.connectors[0])
        expected_inputs.extend(pipeline.inline.connectors[0])
        expected_inputs.extend(pipeline.target.connectors[0])

        expected_outputs = []
        expected_outputs.extend(pipeline.source.connectors[1])
        expected_outputs.extend(pipeline.inline.connectors[1])
        expected_outputs.extend(pipeline.target.connectors[1])

        self.assertCountEqual(expected_inputs, pipeline.connectors[0])
        self.assertCountEqual(expected_outputs, pipeline.connectors[1])


class PipelineValidation(TestCase):
    """Test appropriate errors are raised for an invalid pipeline."""

    def test_orphaned_node(self) -> None:
        """Test a ``OrphanedNodeError`` for an unreachable node"""

        class OrphanedNode(Node):
            def action(self) -> None:
                pass

        class Pipe(Pipeline):
            def __init__(self) -> None:
                self.node = OrphanedNode()
                super().__init__()

        with self.assertRaises(OrphanedNodeError):
            Pipe().validate()

    def test_missing_connection(self) -> None:
        """Test a ``MissingConnectionError`` for an unconnected connector"""

        class Pipe(Pipeline):
            def __init__(self) -> None:
                self.root = MockSource()
                super().__init__()

        with self.assertRaises(MissingConnectionError):
            Pipe().validate()
